import os
import types
from collections.abc import Iterable

import attrs
from rich.traceback import LOCALS_MAX_LENGTH, LOCALS_MAX_STRING, Traceback

from liblaf.grapes import deps


def default_suppress() -> Iterable[str | types.ModuleType]:
    return suppress_converter(
        (deps.try_import("liblaf.cherries"), deps.try_import("pydantic")),
    )


def suppress_converter(
    value: str
    | os.PathLike
    | types.ModuleType
    | Iterable[str | os.PathLike | types.ModuleType | None]
    | None,
) -> Iterable[str | types.ModuleType]:
    if value is None:
        return ()
    if not isinstance(value, Iterable) or isinstance(
        value, (str, os.PathLike, types.ModuleType)
    ):
        value = (value,)
    return [
        val for item in value if (val := _suppress_converter_single(item)) is not None
    ]


def _suppress_converter_single(
    value: str | os.PathLike | types.ModuleType | None,
) -> str | types.ModuleType | None:
    if value is None:
        return None
    if isinstance(value, types.ModuleType):
        return value
    if isinstance(value, str):
        module: types.ModuleType | None = deps.try_import(value)
        if module is not None:
            return module
    return str(value)


@attrs.define
class RichTracebackConfig:
    width: int | None = attrs.field(default=None)
    code_width: int | None = attrs.field(default=None)
    extra_lines: int = attrs.field(default=3)
    theme: str | None = attrs.field(default=None)
    word_wrap: bool = attrs.field(default=False)
    show_locals: bool = attrs.field(default=True)
    locals_max_length: int = attrs.field(default=LOCALS_MAX_LENGTH)
    locals_max_string: int = attrs.field(default=LOCALS_MAX_STRING)
    locals_hide_dunder: bool = attrs.field(default=True)
    locals_hide_sunder: bool = attrs.field(default=False)
    indent_guides: bool = attrs.field(default=True)
    suppress: Iterable[str | types.ModuleType] = attrs.field(
        converter=suppress_converter, factory=default_suppress
    )
    max_frames: int = attrs.field(default=100)

    def from_exception(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: types.TracebackType | None,
        /,
    ) -> Traceback:
        return Traceback.from_exception(
            exc_type,
            exc_value,
            traceback,
            width=self.width,
            code_width=self.code_width,
            extra_lines=self.extra_lines,
            theme=self.theme,
            word_wrap=self.word_wrap,
            show_locals=self.show_locals,
            locals_max_length=self.locals_max_length,
            locals_max_string=self.locals_max_string,
            locals_hide_dunder=self.locals_hide_dunder,
            locals_hide_sunder=self.locals_hide_sunder,
            indent_guides=self.indent_guides,
            suppress=self.suppress,
            max_frames=self.max_frames,
        )



from typing import Protocol

import git
import giturlparse

from liblaf.grapes.typing import PathLike


class GitInfo(Protocol):
    """...

    References:
        1. [nephila/giturlparse: Parse & rewrite git urls (supports GitHub, Bitbucket, Assembla ...)](https://github.com/nephila/giturlparse#exposed-attributes)
    """

    platform: str
    """platform codename"""
    host: str
    """server hostname"""
    resource: str
    """same as `host`"""
    port: str
    """URL port (only if explicitly defined in URL)"""
    protocol: str
    """URL protocol (git, ssh, http/https)"""
    protocols: list[str]
    """list of protocols explicitly defined in URL"""
    user: str
    """repository user"""
    owner: str
    """repository owner (user or organization)"""
    repo: str
    """repository name"""
    name: str
    """same as `repo`"""
    groups: list[str]
    """list of groups - gitlab only"""
    path: str
    """path to file or directory (includes the branch name) - gitlab / github only"""
    path_raw: str
    """raw path starting from the repo name (might include platform keyword) - gitlab / github only"""
    branch: str
    """branch name (when parseable) - gitlab / github only"""
    username: str
    """username from `<username>:<access_token>@<url>` gitlab / github urls"""
    access_token: str
    """access token from `<username>:<access_token>@<url>` gitlab / github urls"""


def info(
    path: PathLike | None = None, *, search_parent_directories: bool = True
) -> GitInfo:
    repo = git.Repo(path=path, search_parent_directories=search_parent_directories)
    url: str = repo.remote().url
    return giturlparse.parse(url)  # pyright: ignore[reportReturnType]



import functools
from collections.abc import Callable
from typing import Any, overload

import attrs

from ._wadler_lindig import pdoc_attrs, pformat


@overload
def auto_repr[T: type](
    cls: T, *, repr: bool | None = None, pdoc: bool | None = None
) -> T: ...
@overload
def auto_repr[T: type](
    *, repr: bool | None = None, pdoc: bool | None = None
) -> Callable[[T], T]: ...
def auto_repr(
    cls: type | None = None,
    *,
    repr: bool | None = None,  # noqa: A002
    pdoc: bool | None = None,
) -> Any:
    if cls is None:
        return functools.partial(auto_repr, repr=repr, pdoc=pdoc)
    if repr or (repr is None and "__repr__" not in cls.__dict__):
        cls.__repr__ = lambda self: pformat(self)  # pyright: ignore[reportAttributeAccessIssue]
    if (pdoc or (pdoc is None and "__pdoc__" not in cls.__dict__)) and attrs.has(cls):
        cls.__pdoc__ = pdoc_attrs  # pyright: ignore[reportAttributeAccessIssue]
    return cls



import math

from liblaf.grapes.error import MatchError

FORMATS: list[tuple[int, float, str]] = [
    (11, 1e-8, "9.99 ns"),
    (10, 1e-7, "99.9 ns"),
    (9, 1e-6, "999. ns"),
    (8, 1e-5, "9.99 µs"),
    (7, 1e-4, "99.9 µs"),
    (6, 1e-3, "999. µs"),
    (5, 1e-2, "9.99 ms"),
    (4, 1e-1, "99.9 ms"),
    (3, 1.0, "999. ms"),
    (2, 1e1, "9.99 s"),
    (1, 60, "99.9 s"),
    (0, 3600.0, "59:59"),
    (0, 86400.0, "23:59:59"),
    (0, math.inf, "1d,23:59:59"),
]


def choose_duration_format(seconds: float) -> str:
    for prec, threshold, fmt in FORMATS:
        if round(seconds, prec) < threshold:
            return fmt
    raise NotImplementedError


def pretty_duration(seconds: float, fmt: str | None = None) -> str:  # noqa: C901, PLR0911, PLR0912
    """.

    Examples:
        >>> pretty_duration(math.nan)
        '?? s'
        >>> pretty_duration(1e-12)
        '0.00 ns'
        >>> pretty_duration(1e-11)
        '0.01 ns'
        >>> pretty_duration(1e-10)
        '0.10 ns'
        >>> pretty_duration(1e-9)
        '1.00 ns'
        >>> pretty_duration(1e-8)
        '10.0 ns'
        >>> pretty_duration(1e-7)
        '100. ns'
        >>> pretty_duration(1e-6)
        '1.00 µs'
        >>> pretty_duration(1e-5)
        '10.0 µs'
        >>> pretty_duration(1e-4)
        '100. µs'
        >>> pretty_duration(1e-3)
        '1.00 ms'
        >>> pretty_duration(1e-2)
        '10.0 ms'
        >>> pretty_duration(1e-1)
        '100. ms'
        >>> pretty_duration(1.0)
        '1.00 s'
        >>> pretty_duration(1e1)
        '10.0 s'
        >>> pretty_duration(1e2)
        '01:40'
        >>> pretty_duration(1e3)
        '16:40'
        >>> pretty_duration(1e4)
        '02:46:40'
        >>> pretty_duration(1e5)
        '1d,03:46:40'
        >>> pretty_duration(1e6)
        '11d,13:46:40'
    """
    if not math.isfinite(seconds):
        return "?? s"
    if fmt is None:
        fmt = choose_duration_format(seconds)
    fmt = fmt.replace("us", "µs")
    match fmt:
        case "9.99 ns":
            return f"{seconds * 1e9:#.2f} ns"
        case "99.9 ns":
            return f"{seconds * 1e9:#.1f} ns"
        case "999. ns":
            return f"{seconds * 1e9:#.0f} ns"
        case "9.99 µs":
            return f"{seconds * 1e6:#.2f} µs"
        case "99.9 µs":
            return f"{seconds * 1e6:#.1f} µs"
        case "999. µs":
            return f"{seconds * 1e6:#.0f} µs"
        case "9.99 ms":
            return f"{seconds * 1e3:#.2f} ms"
        case "99.9 ms":
            return f"{seconds * 1e3:#.1f} ms"
        case "999. ms":
            return f"{seconds * 1e3:#.0f} ms"
        case "9.99 s":
            return f"{seconds:#.2f} s"
        case "99.9 s":
            return f"{seconds:#.1f} s"
        case "999. s":
            return f"{seconds:#.0f} s"
        case "59:59":
            minutes: int
            seconds: int
            seconds = round(seconds)
            minutes, seconds = divmod(seconds, 60)
            return f"{minutes:02d}:{seconds:02d}"
        case "23:59:59":
            hours: int
            minutes: int
            seconds = round(seconds)
            hours, seconds = divmod(seconds, 3600)
            minutes, seconds = divmod(seconds, 60)
            return f"{hours:02d}:{minutes:02d}:{seconds:02d}"
        case "1d,23:59:59":
            days: int
            hours: int
            minutes: int
            seconds = round(seconds)
            days, seconds = divmod(seconds, 86400)
            hours, seconds = divmod(seconds, 3600)
            minutes, seconds = divmod(seconds, 60)
            return f"{days}d,{hours:02d}:{minutes:02d}:{seconds:02d}"
        case _:
            raise MatchError(fmt, "format")



from collections.abc import Mapping, Sequence
from typing import Any

import loguru
import toolz
from loguru import logger

from liblaf.grapes.conf import config
from liblaf.grapes.typing import PathLike

from .filters import FilterLike
from .handlers import file_handler, rich_handler
from .helpers import (
    patch_loguru_get_frame,
    setup_excepthook,
    setup_icecream,
    setup_loguru_intercept,
    setup_unraisablehook,
)


def init(
    *,
    enable_link: bool = True,
    file: PathLike | None = None,
    filter: FilterLike = None,  # noqa: A002
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    intercept: bool = True,
    level: int | str | None = None,
) -> None:
    if file is None:
        file = config.logging.file
    if level is None:
        level = config.logging.level

    if handlers is None:
        handler_config: Mapping[str, Any] = toolz.valfilter(
            lambda v: v is not None, {"filter": filter, "level": level}
        )
        handlers: list[loguru.HandlerConfig] = [
            rich_handler(**handler_config, enable_link=enable_link)
        ]
        if file:
            handlers.append(file_handler(sink=file, **handler_config))

    patch_loguru_get_frame()
    logger.configure(handlers=handlers)
    setup_excepthook()
    setup_icecream()
    setup_unraisablehook()
    if intercept:
        setup_loguru_intercept(level=level)



from pathlib import Path

import pydantic_settings as ps

from ._base import BaseConfig


class ConfigLogging(BaseConfig):
    model_config = ps.SettingsConfigDict(env_prefix="LOG_")
    file: Path | None = None
    level: int | str = "TRACE"



import sys

from loguru import logger


def setup_unraisablehook(level: int | str = "ERROR") -> None:
    def unraisablehook(args: "sys.UnraisableHookArgs", /) -> None:
        if logger is None:  # logger has been cleaned up
            return
        logger.opt(exception=(args.exc_type, args.exc_value, args.exc_traceback)).log(
            level,
            "{err_msg}: {object!r}",
            err_msg=args.err_msg or "Exception ignored in",
            object=args.object,
        )

    sys.unraisablehook = unraisablehook



from __future__ import annotations

import sys
import types
from pathlib import Path

import loguru
from loguru import logger


def setup_excepthook(level: int | str = "CRITICAL") -> None:
    def excepthook(
        exc_type: type[BaseException],
        exc_value: BaseException,
        exc_traceback: types.TracebackType | None,
        /,
    ) -> None:
        logger.opt(exception=(exc_type, exc_value, exc_traceback)).patch(_patcher).log(
            level, exc_value
        )

    sys.excepthook = excepthook


def _patcher(record: loguru.Record) -> None:
    if record["exception"] is None:
        return
    traceback: types.TracebackType | None
    _, _, traceback = record["exception"]
    if traceback is None:
        return
    while traceback.tb_next is not None:
        traceback = traceback.tb_next
    frame: types.FrameType = traceback.tb_frame
    filename: Path = Path(frame.f_code.co_filename)
    record["file"].name = filename.name
    record["file"].path = frame.f_code.co_filename
    record["function"] = frame.f_code.co_name
    record["line"] = frame.f_lineno
    record["module"] = filename.stem
    record["name"] = frame.f_globals.get("__name__", record["name"])



from collections.abc import Callable
from typing import Any

import wrapt

from liblaf.grapes import pretty

from ._base import BaseTimer
from ._utils import set_timer


def timed_callable[**P, T](func: Callable[P, T], timer: BaseTimer) -> Callable[P, T]:
    if timer.name is None:
        timer.name = pretty.pretty_func(func)

    @wrapt.decorator
    def wrapper(
        wrapped: Callable[P, T],
        _instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> T:
        __tracebackhide__ = True
        timer.start()
        try:
            return wrapped(*args, **kwargs)
        finally:
            timer.stop()

    func: Callable[P, T] = wrapper(func)
    set_timer(func, timer)
    return func



import contextlib
import functools
from collections.abc import Callable, Generator, Iterable
from typing import Literal, overload

import joblib
from rich.progress import Progress as RichProgress
from rich.progress import TaskID

from liblaf.grapes import itertools as _it

from ._progress import Progress


@overload
def parallel[T](
    fn: Callable[..., T],
    *iterables: Iterable,
    description: str = "Working...",
    progress: Progress | Literal[False] | None = None,
    return_as: Literal["list"] = "list",
    total: int | None = None,
) -> list[T]: ...
@overload
def parallel[T](
    fn: Callable[..., T],
    *iterables: Iterable,
    description: str = "Working...",
    progress: Progress | Literal[False] | None = None,
    return_as: Literal["generator", "generator_unordered"],
    total: int | None = None,
) -> Generator[T]: ...
def parallel[T](
    fn: Callable[..., T],
    *iterables: Iterable,
    description: str = "Working...",
    progress: Progress | Literal[False] | None = None,
    return_as: Literal["list", "generator", "generator_unordered"] = "list",
    total: int | None = None,
) -> list[T] | Generator[T]:
    for iterable in iterables:
        if total is not None:
            break
        total = _it.len_or_none(iterable)

    parallel = joblib.Parallel(return_as=return_as)

    task_id: TaskID | None = None
    if progress is None:
        progress = Progress()
    progress: Progress | contextlib.nullcontext
    if progress:
        task_id = progress.add_task(description=description, total=total)
    else:
        progress = contextlib.nullcontext()
    parallel.print_progress = functools.partial(
        print_progress, self=parallel, progress=progress, task_id=task_id
    )

    jobs = map(joblib.delayed(fn), *iterables)
    match return_as:
        case "list":
            return as_list(parallel, jobs, progress=progress)
        case "generator" | "generator_unordered":
            return as_generator(parallel, jobs, progress=progress)


def print_progress(
    self: joblib.Parallel,
    progress: RichProgress | contextlib.nullcontext,
    task_id: TaskID | None,
) -> None:
    if isinstance(progress, RichProgress):
        assert task_id is not None
        progress.update(task_id, total=self.n_tasks, completed=self.n_completed_tasks)


def as_list(
    parallel: joblib.Parallel,
    jobs: Iterable,
    *,
    progress: contextlib.AbstractContextManager,
) -> list:
    with progress, parallel:
        return parallel(jobs)  # pyright: ignore[reportReturnType]


def as_generator(
    parallel: joblib.Parallel,
    jobs: Iterable,
    *,
    progress: contextlib.AbstractContextManager,
) -> Generator:
    with progress, parallel:
        yield from parallel(jobs)



from pathlib import Path

from rich.style import Style
from rich.text import Text

from liblaf.grapes.typing import PathLike


def rich_location(
    name: str | None,
    function: str | None,
    line: int | None,
    file: PathLike | None = None,
    *,
    enable_link: bool = True,
    width: int | None = None,
) -> Text:
    name = name or "<unknown>"
    function = function or "<unknown>"
    line = line or 0
    file: Path | None = Path(file) if file is not None else None
    func_line: str = f":{function}:{line}"
    if width and len(name) + len(func_line) > width:
        name = name[: width - len(func_line) - 1] + "…"
    if enable_link and file is not None and file.exists():
        return Text(
            f"{name}:{function}:{line}", style=Style(link=f"{file.as_uri()}#{line}")
        )
    return Text(f"{name}:{function}:{line}")



import functools
from collections.abc import Buffer, Callable
from pathlib import Path
from typing import Any, Literal, overload

import attrs
import msgspec

from liblaf.grapes.typing import PathLike

from ._decode import DecHook, PydanticModelValidateOptions, dec_hook
from ._encode import EncHook, PydanticModelDumpOptions, enc_hook

type Decoder = Callable
type Encoder = Callable


@attrs.define
class Serde:
    decoder: Decoder
    encoder: Encoder

    @overload
    def decode(
        self,
        buf: Buffer | str,
        /,
        *,
        dec_hook: DecHook | None = ...,
        pydantic: PydanticModelValidateOptions | None = None,
        strict: bool = True,
    ) -> Any: ...
    @overload
    def decode[T](
        self,
        buf: Buffer | str,
        /,
        *,
        dec_hook: DecHook | None = ...,
        pydantic: PydanticModelValidateOptions | None = None,
        strict: bool = True,
        type: type[T],
    ) -> T: ...
    @overload
    def decode[T](
        self,
        buf: Buffer | str,
        /,
        *,
        dec_hook: DecHook | None = ...,
        pydantic: PydanticModelValidateOptions | None = None,
        strict: bool = True,
        type: Any,
    ) -> Any: ...
    def decode(self, buf: Buffer | str, /, **kwargs) -> Any:
        if "dec_hook" not in kwargs:
            kwargs["dec_hook"] = functools.partial(
                dec_hook, pydantic_options=kwargs.pop("pydantic", None)
            )
        return self.decoder(buf, **kwargs)

    @overload
    def encode(  # pyright: ignore[reportInconsistentOverload]
        self,
        obj: Any,
        /,
        *,
        enc_hook: EncHook | None = ...,
        order: Literal["deterministic", "sorted"] | None = None,
        pydantic: PydanticModelDumpOptions | None = None,
    ) -> bytes: ...
    def encode(self, obj: Any, /, **kwargs) -> bytes:
        if "enc_hook" not in kwargs:
            kwargs["enc_hook"] = functools.partial(
                enc_hook, pydantic_options=kwargs.pop("pydantic", None)
            )
        return self.encoder(obj, **kwargs)

    @overload
    def load(
        self,
        path: PathLike,
        /,
        *,
        dec_hook: DecHook | None = ...,
        pydantic: PydanticModelValidateOptions | None = None,
        strict: bool = True,
    ) -> Any: ...
    @overload
    def load[T](
        self,
        path: PathLike,
        /,
        *,
        dec_hook: DecHook | None = ...,
        pydantic: PydanticModelValidateOptions | None = None,
        strict: bool = True,
        type: type[T],
    ) -> T: ...
    @overload
    def load[T](
        self,
        path: PathLike,
        /,
        *,
        dec_hook: DecHook | None = ...,
        pydantic: PydanticModelValidateOptions | None = None,
        strict: bool = True,
        type: Any,
    ) -> Any: ...
    def load(self, path: PathLike, /, **kwargs) -> Any:
        path = Path(path)
        return self.decode(path.read_bytes(), **kwargs)

    @overload
    def save(  # pyright: ignore[reportInconsistentOverload]
        self,
        path: PathLike,
        obj: Any,
        /,
        *,
        enc_hook: EncHook | None = ...,
        order: Literal["deterministic", "sorted"] | None = None,
        pydantic: PydanticModelDumpOptions | None = None,
    ) -> None: ...
    def save(self, path: PathLike, obj: Any, /, **kwargs) -> None:
        path = Path(path)
        path.write_bytes(self.encode(obj, **kwargs))


json = Serde(decoder=msgspec.json.decode, encoder=msgspec.json.encode)
toml = Serde(decoder=msgspec.toml.decode, encoder=msgspec.toml.encode)
yaml = Serde(decoder=msgspec.yaml.decode, encoder=msgspec.yaml.encode)



import types
import unittest.mock
from collections.abc import Callable, Generator
from pathlib import Path

from loguru._get_frame import load_get_frame_function

from liblaf.grapes.conf import config

from ._traceback import _validate_suppress

_get_frame_original: Callable[[int], types.FrameType | None] = load_get_frame_function()


def _get_frame(depth: int = 0, /) -> types.FrameType | None:
    __tracebackhide__ = True
    frame: types.FrameType | None = _get_frame_original(0)
    while frame is not None:
        if not _should_hide(frame):
            if depth <= 0:
                return frame
            depth -= 1
        frame = frame.f_back
    if frame is None:
        msg = "call stack is not deep enough"
        raise ValueError(msg)
    return frame


def _should_hide(frame: types.FrameType) -> bool:
    if frame.f_locals.get("__tracebackhide__"):
        return True
    file: Path = Path(frame.f_code.co_filename)
    suppressed: bool = any(file.is_relative_to(path) for path in _get_suppress_paths())
    return suppressed


def _get_suppress_paths() -> Generator[Path]:
    suppress = _validate_suppress(config.traceback.suppress)
    for item in suppress:
        if isinstance(item, str):
            yield Path(item)
        else:
            file: str | None = getattr(item, "__file__", None)
            if file is None:
                continue
            yield Path(file).parent


def patch_loguru_get_frame(
    new: Callable[[int], types.FrameType | None] = _get_frame,
) -> None:
    # ! dirty hack
    patcher = unittest.mock.patch("loguru._logger.get_frame", new)
    patcher.start()



from ._array import array_kind
from ._types import ClassInfo, PathLike
from ._utils import clone_param_spec, clone_signature

__all__ = ["ClassInfo", "PathLike", "array_kind", "clone_param_spec", "clone_signature"]



import math
import statistics
from collections.abc import Sequence

import autoregistry

from liblaf.grapes import pretty

type StatisticName = str


STATISTICS_REGISTRY = autoregistry.Registry(prefix="_compute_")


STATISTICS_REGISTRY["max"] = max
STATISTICS_REGISTRY["mean"] = statistics.mean
STATISTICS_REGISTRY["median"] = statistics.median
STATISTICS_REGISTRY["min"] = min
STATISTICS_REGISTRY["stdev"] = statistics.stdev
STATISTICS_REGISTRY["total"] = sum


def compute_statistic(series: Sequence[float], stat_name: StatisticName) -> float:
    try:
        return STATISTICS_REGISTRY[stat_name](series)
    except (ValueError, statistics.StatisticsError):
        return math.nan


def pretty_statistic(
    series: Sequence[float], stat_name: StatisticName
) -> tuple[str, str]:
    match stat_name:
        case "mean+stdev":
            mean: float = compute_statistic(series, "mean")
            stdev: float = compute_statistic(series, "stdev")
            mean_str: str = pretty.pretty_duration(mean)
            stdev_str: str = pretty.pretty_duration(stdev)
            fmt = "{} ± {}"
            pretty_name: str = fmt.format("mean", "σ")  # noqa: RUF001
            pretty_value: str = fmt.format(mean_str, stdev_str)
            return pretty_name, pretty_value
        case "range":
            minimum: float = compute_statistic(series, "min")
            maximum: float = compute_statistic(series, "max")
            min_str: str = pretty.pretty_duration(minimum)
            max_str: str = pretty.pretty_duration(maximum)
            fmt = "{} … {}"
            pretty_name: str = fmt.format("min", "max")
            pretty_value: str = fmt.format(min_str, max_str)
            return pretty_name, pretty_value
        case stat_name:
            value: float = compute_statistic(series, stat_name)
            return stat_name, pretty.pretty_duration(value)



import types
from collections.abc import Generator, Sequence

import attrs
import loguru
from rich.console import Console, RenderableType

from liblaf.grapes import pretty
from liblaf.grapes.logging.helpers import rich_traceback

from .columns import (
    RichSinkColumn,
    RichSinkColumnElapsed,
    RichSinkColumnLevel,
    RichSinkColumnLocation,
    RichSinkColumnMessage,
)


def default_columns(*, enable_link: bool = True) -> Sequence[RichSinkColumn]:
    return [
        RichSinkColumnElapsed(),
        RichSinkColumnLevel(),
        RichSinkColumnLocation(enable_link=enable_link),
        RichSinkColumnMessage(),
    ]


def default_console() -> Console:
    return pretty.get_console(stderr=True)


@attrs.define
class RichSink:
    columns: Sequence[RichSinkColumn] = attrs.field(
        converter=attrs.converters.default_if_none(factory=default_columns),
        factory=default_columns,
    )
    console: Console = attrs.field(
        converter=attrs.converters.default_if_none(factory=default_console),
        factory=default_console,
    )

    def __init__(
        self,
        columns: Sequence[RichSinkColumn] | None = None,
        console: Console | None = None,
        *,
        enable_link: bool = True,
    ) -> None:
        if columns is None:
            columns = default_columns(enable_link=enable_link)
        self.__attrs_init__(columns=columns, console=console)  # pyright: ignore[reportAttributeAccessIssue]

    def __call__(self, message: "loguru.Message", /) -> None:
        record: loguru.Record = message.record
        # TODO: `console.print()` is slow
        self.console.print(
            *self.render(record), overflow="ignore", no_wrap=True, crop=False
        )
        if (excpetion := self.render_exception(record)) is not None:
            self.console.print(excpetion)

    def render(self, record: "loguru.Record", /) -> Generator[RenderableType]:
        for column in self.columns:
            yield column.render(record)

    def render_exception(self, record: "loguru.Record", /) -> RenderableType | None:
        exception: loguru.RecordException | None = record["exception"]
        if exception is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = exception
        if exc_type is None or exc_value is None:
            return None
        return rich_traceback(exc_type, exc_value, traceback)



from pathlib import Path
from typing import Unpack

import loguru

from liblaf.grapes.conf import config
from liblaf.grapes.logging.filters import new_filter
from liblaf.grapes.logging.helpers import new_format


def file_handler(
    **kwargs: Unpack["loguru.FileHandlerConfig"],
) -> "loguru.FileHandlerConfig":
    if "sink" not in kwargs:
        kwargs["sink"] = config.logging.file or Path("run.log")
    if "format" not in kwargs:
        kwargs["format"] = new_format()
    kwargs["filter"] = new_filter(kwargs.get("filter"))
    kwargs.setdefault("mode", "w")
    return kwargs



import inspect
import itertools
import logging
import types
from collections.abc import Iterable

import loguru._logger
from loguru import logger


def clear_stdlib_handlers() -> None:
    for std_logger in logging.root.manager.loggerDict.values():
        if isinstance(std_logger, logging.PlaceHolder):
            continue
        std_logger.handlers.clear()
        std_logger.propagate = True


def setup_loguru_intercept(
    level: int | str = logging.NOTSET, modules: Iterable[str] = ()
) -> None:
    """Logs to loguru from Python logging module.

    References:
        1. [Entirely compatible with standard logging](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
        2. [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py)
    """
    core: loguru._logger.Core = logger._core  # pyright: ignore[reportAttributeAccessIssue] # noqa: SLF001
    for loguru_level in core.levels.values():
        logging.addLevelName(loguru_level.no, loguru_level.name)
    if isinstance(level, str):
        level = logger.level(level).no
    logging.basicConfig(handlers=[InterceptHandler()], level=level)
    clear_stdlib_handlers()
    for logger_name in itertools.chain(("",), modules):
        mod_logger: logging.Logger = logging.getLogger(logger_name)
        mod_logger.handlers = [InterceptHandler()]
        mod_logger.propagate = False


class InterceptHandler(logging.Handler):
    """Logs to loguru from Python logging module.

    References:
        1. [Entirely compatible with standard logging](https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
        2. [loguru-logging-intercept/loguru_logging_intercept.py at f358b75ef4162ea903bf7a3298c22b1be83110da · MatthewScholefield/loguru-logging-intercept](https://github.com/MatthewScholefield/loguru-logging-intercept/blob/f358b75ef4162ea903bf7a3298c22b1be83110da/loguru_logging_intercept.py)
    """

    def emit(self, record: logging.LogRecord) -> None:
        if logger is None:  # logger has been cleaned up
            return

        # Get corresponding Loguru level if it exists.
        level: str | int
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame: types.FrameType | None = inspect.currentframe()
        depth: int = 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )



from __future__ import annotations

from collections.abc import Hashable

import attrs
import loguru


@attrs.define
class FilterOnce:
    _history: set[Hashable] = attrs.field(factory=set, init=False)

    def __call__(self, record: loguru.Record) -> bool:
        if not record["extra"].get("once", False):
            return True
        record_hash: Hashable = self._hash_record(record)
        if record_hash in self._history:
            return False
        self._history.add(record_hash)
        return True

    def _hash_record(self, record: loguru.Record) -> Hashable:
        return (
            record["function"],
            record["level"].no,
            record["line"],
            record["message"],
            record["name"],
        )



from ._as_iterable import as_iterable
from ._as_sequence import as_sequence
from ._deep_merge import deep_merge
from ._first_not_none import first_not_none
from ._len_or_none import len_or_none

__all__ = ["as_iterable", "as_sequence", "deep_merge", "first_not_none", "len_or_none"]



from ._add_level import add_level
from ._excepthook import setup_excepthook
from ._format import new_format
from ._get_frame import patch_loguru_get_frame
from ._icecream import setup_icecream
from ._std_intercept import (
    InterceptHandler,
    clear_stdlib_handlers,
    setup_loguru_intercept,
)
from ._traceback import rich_traceback
from ._unraisablehook import setup_unraisablehook

__all__ = [
    "InterceptHandler",
    "add_level",
    "clear_stdlib_handlers",
    "new_format",
    "patch_loguru_get_frame",
    "rich_traceback",
    "setup_excepthook",
    "setup_icecream",
    "setup_loguru_intercept",
    "setup_unraisablehook",
]



from . import typing
from ._by_name import FilterByName
from ._by_version import FilterByVersion
from ._composite import CompositeFilter
from ._factory import new_filter
from ._once import FilterOnce
from ._utils import as_level_no_dict, get_level_no, get_parent
from .typing import FilterLike

__all__ = [
    "CompositeFilter",
    "FilterByName",
    "FilterByVersion",
    "FilterLike",
    "FilterOnce",
    "as_level_no_dict",
    "get_level_no",
    "get_parent",
    "new_filter",
    "typing",
]



import inspect
from collections.abc import Callable, Mapping, Sequence
from typing import Any

import attrs
import wadler_lindig as wl

from ._utils import get_name
from ._wadler_lindig import pformat


@attrs.define
class PrettyCall:
    func: Callable
    args: Sequence[Any] = ()
    kwargs: Mapping[str, Any] = {}

    def __repr__(self) -> str:
        return pformat(self)

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc:
        name: str = get_name(self.func)
        params: list[wl.AbstractDoc] = [wl.pdoc(arg, **kwargs) for arg in self.args]
        params += wl.named_objs(self.kwargs.items(), **kwargs)
        return wl.bracketed(
            begin=wl.TextDoc(name) + wl.TextDoc("("),
            docs=params,
            sep=wl.comma,
            end=wl.TextDoc(")"),
            indent=kwargs["indent"],
        )


def pretty_call(
    func: Callable,
    args: Sequence[Any] = (),
    kwargs: Mapping[str, Any] = {},
    **wl_kwargs,
) -> str:
    func = inspect.unwrap(func)
    args, kwargs = _bind_safe(func, args, kwargs)
    return pformat(PrettyCall(func, args, kwargs), **wl_kwargs)


def _bind_safe(
    func: Callable, /, args: Sequence[Any], kwargs: Mapping[str, Any]
) -> tuple[Sequence[Any], Mapping[str, Any]]:
    try:
        signature: inspect.Signature = inspect.signature(func)
        arguments: inspect.BoundArguments = signature.bind(*args, **kwargs)
    except TypeError:
        return args, kwargs
    else:
        return arguments.args, arguments.kwargs



from __future__ import annotations

import sys
import types
from typing import override

import attrs
import loguru
from rich.console import RenderableType
from rich.text import Text

from liblaf.grapes import pretty

from ._abc import RichSinkColumn


@attrs.define
class RichSinkColumnLocation(RichSinkColumn):
    abbr_name: bool = True
    enable_link: bool = True
    width: int | None = None

    @override  # impl RichSinkColumn
    def render(self, record: loguru.Record, /) -> RenderableType:
        name: str | None = self.render_name(record)
        location: Text = pretty.rich_location(
            name=name,
            function=record["function"],
            line=record["line"],
            file=record["file"].path,
            enable_link=self.enable_link,
            width=self.width,
        )
        location.style = "log.path"
        return location

    def render_name(self, record: loguru.Record) -> str | None:
        name: str | None = record["name"]
        if name is None:
            return None
        if not self.abbr_name:
            return name
        function: str = record["function"]
        start: int = 0
        while True:
            index: int = name.find(".", start)
            if index < 0:
                return name
            module_name: str = name[:index]
            module: types.ModuleType | None = sys.modules.get(module_name)
            if module is None:
                return name
            if hasattr(module, function):
                return module_name
            start = index + 1



import pydantic

from ._base import BaseConfig
from ._joblib import ConfigJoblib
from ._logging import ConfigLogging
from ._pretty import ConfigPretty
from ._traceback import ConfigTraceback


class Config(BaseConfig):
    joblib: ConfigJoblib = pydantic.Field(default_factory=ConfigJoblib)
    logging: ConfigLogging = pydantic.Field(default_factory=ConfigLogging)
    pretty: ConfigPretty = pydantic.Field(default_factory=ConfigPretty)
    traceback: ConfigTraceback = pydantic.Field(default_factory=ConfigTraceback)


config = Config()



from __future__ import annotations

import attrs
import loguru

from ._utils import as_level_no_dict, get_parent


@attrs.define
class FilterByName:
    _levels: dict[str, int] = attrs.field(
        default=None,
        converter=[
            attrs.converters.default_if_none(factory=lambda: {"__main__": "TRACE"}),
            as_level_no_dict,
        ],
    )  # pyright: ignore[reportAssignmentType]

    def __call__(self, record: loguru.Record) -> bool:
        level: int | None = self.get_level(record["name"])
        if level is None:
            return True
        return record["level"].no >= level

    def get_level(self, name: str | None) -> int | None:
        while True:
            if not name:
                return None
            level: int | None = self._levels.get(name)
            if level is not None:
                return level
            name = get_parent(name)



import loguru._logger
from loguru import logger


def add_level(
    name: str, no: int | None = None, color: str | None = None, icon: str | None = None
) -> "loguru.Level":
    core: loguru._logger.Core = logger._core  # pyright: ignore[reportAttributeAccessIssue] # noqa: SLF001
    if name in core.levels and no == logger.level(name).no:
        no = None  # skip update severity no
    return logger.level(name=name, no=no, color=color, icon=icon)



import datetime

import attrs
import loguru
from rich.console import RenderableType
from rich.text import Text

from ._abc import RichSinkColumn


@attrs.define
class RichSinkColumnTime(RichSinkColumn):
    fmt: str = "%Y-%m-%dT%H:%M:%S.%f"

    def render(self, record: "loguru.Record", /) -> RenderableType:
        time: datetime.datetime = record["time"]
        return Text(f"{time:{self.fmt}}", style="log.time")



import pydantic
import pydantic_settings as ps

from ._base import BaseConfig


class ConfigTraceback(BaseConfig):
    model_config = ps.SettingsConfigDict(env_prefix="TRACEBACK_")
    width: int | None = None
    show_locals: bool = True
    locals_hide_sunder: bool = True
    suppress: list[str] = pydantic.Field(
        default=["comet_ml", "liblaf.cherries", "pydantic", "rich.progress"]
    )



import datetime
import functools
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Literal, Protocol, TypedDict, overload

import cytoolz as toolz
import joblib
import wrapt

from liblaf.grapes.conf import config

from ._wrapt import wrapt_setattr


class MemorizedFunc[**P, T](Protocol):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...


class Metadata(TypedDict): ...


@overload
def cache[**P, T](
    func: Callable[P, T],
    /,
    *,
    memory: joblib.Memory | None = ...,
    # memory.cache() params
    ignore: list[str] | None = ...,
    verbose: int | None = ...,
    mmap_mode: Literal["r+", "r", "w+", "c"] | None = ...,
    cache_validation_callback: Callable[[Metadata], bool] | None = ...,
    # memory.reduce_size() params
    bytes_limit: int | str | None = ...,
    items_limit: int | None = ...,
    age_limit: datetime.timedelta | None = ...,
) -> MemorizedFunc[P, T]: ...
@overload
def cache[**P, T](
    *,
    memory: joblib.Memory | None = None,
    # memory.cache() params
    ignore: list[str] | None = ...,
    verbose: int | None = ...,
    mmap_mode: Literal["r+", "r", "w+", "c"] | None = ...,
    cache_validation_callback: Callable[[Metadata], bool] | None = ...,
    # memory.reduce_size() params
    bytes_limit: int | str | None = ...,
    items_limit: int | None = ...,
    age_limit: datetime.timedelta | None = ...,
) -> Callable[[Callable[P, T]], MemorizedFunc[P, T]]: ...
def cache(func: Callable | None = None, /, **kwargs: Any) -> Any:
    if func is None:
        return functools.partial(cache, **kwargs)
    memory: joblib.Memory | None = kwargs.pop("memory", None)
    if memory is None:
        memory = new_memory()
    cache_kwargs: dict[str, Any] = _filter_keys(
        kwargs, ("ignore", "verbose", "mmap_mode", "cache_validation_callback")
    )
    reduce_size_kwargs: dict[str, Any] = _filter_keys(
        kwargs, ("bytes_limit", "items_limit", "age_limit")
    )
    reduce_size_kwargs.setdefault("bytes_limit", config.joblib.memory.bytes_limit)

    @wrapt.function_wrapper
    def wrapper[**P, T](
        wrapped: Callable[P, T], _instance: Any, args: tuple, kwargs: dict[str, Any]
    ) -> T:
        result: Any = wrapped(*args, **kwargs)
        memory.reduce_size(**reduce_size_kwargs)
        return result

    func = memory.cache(func, **cache_kwargs)
    func = wrapper(func)
    wrapt_setattr(func, "memory", memory)
    return func


@functools.cache
def new_memory() -> joblib.Memory:
    return joblib.Memory(location=config.joblib.memory.location)


def _filter_keys[KT, VT](mapping: Mapping[KT, VT], keys: Iterable[KT]) -> dict[KT, VT]:
    return toolz.keyfilter(lambda k: k in keys, mapping)



from . import filters, handlers, helpers, sink
from ._depth_tracker import helper
from ._init import init
from .filters import (
    CompositeFilter,
    FilterByName,
    FilterByVersion,
    FilterLike,
    FilterOnce,
    as_level_no_dict,
    get_level_no,
    new_filter,
)
from .handlers import file_handler, rich_handler
from .helpers import (
    InterceptHandler,
    add_level,
    clear_stdlib_handlers,
    new_format,
    patch_loguru_get_frame,
    rich_traceback,
    setup_excepthook,
    setup_icecream,
    setup_loguru_intercept,
    setup_unraisablehook,
)
from .sink import (
    RichSink,
    RichSinkColumn,
    RichSinkColumnElapsed,
    RichSinkColumnLevel,
    RichSinkColumnLocation,
    RichSinkColumnMessage,
    RichSinkColumnTime,
    default_columns,
    default_console,
)

__all__ = [
    "CompositeFilter",
    "FilterByName",
    "FilterByVersion",
    "FilterLike",
    "FilterOnce",
    "InterceptHandler",
    "RichSink",
    "RichSinkColumn",
    "RichSinkColumnElapsed",
    "RichSinkColumnLevel",
    "RichSinkColumnLocation",
    "RichSinkColumnMessage",
    "RichSinkColumnTime",
    "add_level",
    "as_level_no_dict",
    "clear_stdlib_handlers",
    "default_columns",
    "default_console",
    "file_handler",
    "filters",
    "get_level_no",
    "handlers",
    "helper",
    "helpers",
    "init",
    "new_filter",
    "new_format",
    "patch_loguru_get_frame",
    "rich_handler",
    "rich_traceback",
    "setup_excepthook",
    "setup_icecream",
    "setup_loguru_intercept",
    "setup_unraisablehook",
    "sink",
]



from __future__ import annotations

from collections.abc import Mapping

from ._composite import CompositeFilter
from .typing import FilterLike


def new_filter(f: FilterLike | None = None, /) -> FilterLike:
    if f is None:
        return CompositeFilter()
    if isinstance(f, Mapping):
        return CompositeFilter(f)  # pyright: ignore[reportArgumentType]
    return f



from collections.abc import Iterable, Sequence

from ._clock import ClockName
from ._statistics import StatisticName

DEFAULT_CLOCKS: Sequence[ClockName] = ("perf",)


LOG_RECORD_DEFAULT_INDEX: int = -1
LOG_RECORD_DEFAULT_LEVEL: int | str = "DEBUG"
LOG_RECORD_DEFAULT_THRESHOLD_SEC: float | None = 0.02


LOG_SUMMARY_DEFAULT_LEVEL: int | str = "INFO"
LOG_SUMMARY_DEFAULT_STATISTICS: Iterable[StatisticName] = (
    "total",
    "mean+stdev",
    "range",
    "median",
)



import functools
from collections.abc import Callable
from typing import Any, TypedDict, Unpack

import attrs
import cytoolz as toolz
import wadler_lindig as wl

from liblaf.grapes.conf import config
from liblaf.grapes.typing import array_kind

from ._console import get_console

UNINITIALIZED = wl.TextDoc("<uninitialized>")


class WadlerLindigOptions(TypedDict, total=False):
    width: int | None
    indent: int
    short_arrays: bool
    custom: Callable[[Any], wl.AbstractDoc | None]
    hide_defaults: bool
    show_type_module: bool
    show_dataclass_module: bool
    show_function_module: bool
    respect_pdoc: bool
    short_arrays_threshold: int


def pdoc_attrs(self: Any, **kwargs: Unpack[WadlerLindigOptions]) -> wl.AbstractDoc:
    """.

    References:
        1. <https://github.com/patrick-kidger/wadler_lindig/blob/0226340d56f0c18e10cd4d375cf7ea25818359b8/wadler_lindig/_definitions.py#L308-L326>
    """
    kwargs: WadlerLindigOptions = _make_kwargs(kwargs)
    cls: type = type(self)
    objs: list[tuple[str, Any]] = []
    for field in attrs.fields(cls):
        field: attrs.Attribute
        if not field.repr:
            continue
        value: Any = getattr(self, field.name, UNINITIALIZED)
        if kwargs.get("hide_defaults", True) and value is field.default:
            continue
        objs.append((field.name, value))
    name_kwargs: dict[str, Any] = toolz.assoc(
        kwargs, "show_type_module", kwargs.get("show_dataclass_module", False)
    )
    return wl.bracketed(
        begin=wl.pdoc(cls, **name_kwargs) + wl.TextDoc("("),
        docs=wl.named_objs(objs, **kwargs),
        sep=wl.comma,
        end=wl.TextDoc(")"),
        indent=kwargs.get("indent", 2),
    )


@functools.singledispatch
def pdoc_custom(
    obj: Any, **kwargs: Unpack[WadlerLindigOptions]
) -> wl.AbstractDoc | None:
    if hasattr(obj, "__pdoc__"):
        return None
    if attrs.has(type(obj)):
        return pdoc_attrs(obj, **kwargs)
    if array_kind(obj):
        if kwargs.get("short_arrays") is None:
            kwargs["short_arrays"] = obj.size > kwargs.get(
                "short_arrays_threshold", 100
            )
        return wl.pdoc(obj, **kwargs)
    return None


@functools.singledispatch
def pformat(obj: Any, **kwargs: Unpack[WadlerLindigOptions]) -> str:
    kwargs: WadlerLindigOptions = _make_kwargs(kwargs)
    if kwargs.get("width") is None:
        kwargs["width"] = get_console(stderr=True).width
    if kwargs.get("custom") is None:
        kwargs["custom"] = functools.partial(pdoc_custom, **kwargs)
    return wl.pformat(obj, **kwargs)  # pyright: ignore[reportArgumentType]


def _make_kwargs(kwargs: WadlerLindigOptions) -> WadlerLindigOptions:
    kwargs: WadlerLindigOptions = toolz.merge(config.pretty.model_dump(), kwargs)
    return kwargs



from ._decode import DecHook, PydanticModelValidateOptions, dec_hook
from ._encode import EncHook, PydanticModelDumpOptions, enc_hook
from ._load import load
from ._save import save
from ._serde import Serde, json, toml, yaml

__all__ = [
    "DecHook",
    "EncHook",
    "PydanticModelDumpOptions",
    "PydanticModelValidateOptions",
    "Serde",
    "dec_hook",
    "enc_hook",
    "json",
    "load",
    "save",
    "toml",
    "yaml",
]



from typing import override

import loguru
from rich.console import RenderableType
from rich.text import Text

from ._abc import RichSinkColumn


class RichSinkColumnLevel(RichSinkColumn):
    width: int = 1

    @override  # impl RichSinkColumn
    def render(self, record: "loguru.Record", /) -> RenderableType:
        level: str = record["level"].name
        return Text(f"{level:<.{self.width}}", style=f"logging.level.{level.lower()}")



from ._base import BaseConfig, BaseModel
from ._config import Config, config
from ._joblib import ConfigJoblib, ConfigJoblibMemory
from ._logging import ConfigLogging
from ._pretty import ConfigPretty
from ._traceback import ConfigTraceback

__all__ = [
    "BaseConfig",
    "BaseModel",
    "Config",
    "ConfigJoblib",
    "ConfigJoblibMemory",
    "ConfigLogging",
    "ConfigPretty",
    "ConfigTraceback",
    "config",
]



from loguru import logger

from liblaf.grapes import pretty

from ._add_level import add_level


def setup_icecream(prefix: str = "") -> None:
    try:
        from icecream import ic
    except ImportError:
        return

    add_level(name="ICECREAM", no=15, color="<magenta><bold>", icon="🍦")

    def output_function(s: str) -> None:
        logger.opt(depth=2).log("ICECREAM", s)

    ic.configureOutput(
        prefix=prefix,
        argToStringFunction=pretty.pformat,
        outputFunction=output_function,
    )



import sys
import types
import unittest.mock
from collections.abc import Generator, Iterable
from typing import TypedDict, Unpack, cast

import cytoolz as toolz
from rich.text import Text
from rich.traceback import Traceback

from liblaf.grapes import pretty
from liblaf.grapes.conf import config


class TracebackOptions(TypedDict, total=False):
    width: int | None
    show_locals: bool
    suppress: Iterable[str | types.ModuleType]


def rich_traceback(
    exc_type: type[BaseException],
    exc_value: BaseException,
    traceback: types.TracebackType | None,
    /,
    **kwargs: Unpack[TracebackOptions],
) -> Traceback:
    kwargs = toolz.merge(config.traceback.model_dump(), kwargs)
    kwargs = cast("TracebackOptions", kwargs)
    if kwargs.get("width") is None:
        kwargs["width"] = pretty.get_console(stderr=True).width
    if suppress := kwargs.get("suppress"):
        kwargs["suppress"] = _validate_suppress(suppress)
    traceback = _filter_traceback(traceback)
    # ? dirty hack to avoid long `repr()` output
    # ref: <https://github.com/Textualize/rich/discussions/3774>
    with unittest.mock.patch("rich.pretty.repr", new=pretty.pformat):
        rich_tb: Traceback = Traceback.from_exception(
            exc_type, exc_value, traceback, **kwargs
        )
    # ? dirty hack to support ANSI in exception messages
    for stack in rich_tb.trace.stacks:
        if pretty.has_ansi(stack.exc_value):
            stack.exc_value = Text.from_ansi(stack.exc_value)  # pyright: ignore[reportAttributeAccessIssue]
    return rich_tb


def _filter_traceback(
    traceback: types.TracebackType | None,
) -> types.TracebackType | None:
    if traceback is None:
        return None
    if traceback.tb_frame.f_locals.get("__tracebackhide__"):
        return _filter_traceback(traceback.tb_next)
    traceback.tb_next = _filter_traceback(traceback.tb_next)
    return traceback


def _validate_suppress(
    suppress: Iterable[str | types.ModuleType],
) -> Generator[str | types.ModuleType]:
    for item in suppress:
        yield _validate_suppress_single(item)


def _validate_suppress_single(
    suppress: str | types.ModuleType,
) -> str | types.ModuleType:
    if isinstance(suppress, types.ModuleType):
        return suppress
    return sys.modules.get(suppress, suppress)



import attrs

from liblaf.grapes.sentinel import NOP

from . import callback
from ._clock import clock
from ._timings import Callback, Timings


@attrs.define
class BaseTimer(Timings):
    cb_start: Callback | None = None
    cb_stop: Callback | None = callback.log_record
    cb_finish: Callback | None = callback.log_summary

    def __bool__(self) -> bool:
        return True

    def start(self) -> None:
        __tracebackhide__ = True
        for clock_name in self.clocks:
            self._start_time[clock_name] = clock(clock_name)
        self._stop_time.clear()
        if self.cb_start is not None and self.cb_start is not NOP:
            self.cb_start(self)

    def stop(self) -> None:
        __tracebackhide__ = True
        for clock_name in self.clocks:
            stop_time: float = clock(clock_name)
            self._stop_time[clock_name] = stop_time
            self.timings[clock_name].append(stop_time - self._start_time[clock_name])
        if self.cb_stop is not None and self.cb_stop is not NOP:
            self.cb_stop(self)

    def finish(self) -> None:
        __tracebackhide__ = True
        if self.cb_finish is not None and self.cb_finish is not NOP:
            self.cb_finish(self)



from . import (
    conf,
    env,
    error,
    functools,
    git,
    itertools,
    logging,
    pretty,
    sentinel,
    serde,
    timing,
    tqdm,
    typing,
)
from ._version import __version__, __version_tuple__, version, version_tuple
from .conf import BaseConfig, BaseModel, Config, config
from .deps import has_module, optional_imports, try_import
from .error import (
    DispatchLookupError,
    MatchError,
    TodoError,
    UnreachableError,
    todo,
    unreachable,
)
from .functools import MemorizedFunc, cache, wrapt_getattr, wrapt_setattr
from .itertools import as_iterable, as_sequence, deep_merge, first_not_none, len_or_none
from .pretty import (
    WadlerLindigOptions,
    auto_repr,
    choose_duration_format,
    get_console,
    has_ansi,
    pdoc_attrs,
    pdoc_custom,
    pformat,
    pretty_call,
    pretty_duration,
    pretty_func,
    pretty_throughput,
    rich_location,
)
from .sentinel import MISSING, NOP, nop
from .serde import (
    DecHook,
    EncHook,
    PydanticModelDumpOptions,
    PydanticModelValidateOptions,
    Serde,
    dec_hook,
    enc_hook,
    json,
    load,
    save,
    toml,
    yaml,
)
from .timing import BaseTimer, ClockName, Timer, Timings, clock, get_timer, timer
from .tqdm import Progress, RateColumn, parallel, track
from .typing import ClassInfo, PathLike, array_kind, clone_param_spec, clone_signature

__all__ = [
    "MISSING",
    "NOP",
    "BaseConfig",
    "BaseModel",
    "BaseTimer",
    "ClassInfo",
    "ClockName",
    "Config",
    "DecHook",
    "DispatchLookupError",
    "EncHook",
    "MatchError",
    "MemorizedFunc",
    "PathLike",
    "Progress",
    "PydanticModelDumpOptions",
    "PydanticModelValidateOptions",
    "RateColumn",
    "Serde",
    "Timer",
    "Timings",
    "TodoError",
    "UnreachableError",
    "WadlerLindigOptions",
    "__version__",
    "__version_tuple__",
    "array_kind",
    "as_iterable",
    "as_sequence",
    "auto_repr",
    "cache",
    "choose_duration_format",
    "clock",
    "clone_param_spec",
    "clone_signature",
    "conf",
    "config",
    "dec_hook",
    "deep_merge",
    "enc_hook",
    "env",
    "error",
    "first_not_none",
    "functools",
    "get_console",
    "get_timer",
    "git",
    "has_ansi",
    "has_module",
    "itertools",
    "json",
    "len_or_none",
    "load",
    "logging",
    "nop",
    "optional_imports",
    "parallel",
    "pdoc_attrs",
    "pdoc_custom",
    "pformat",
    "pretty",
    "pretty_call",
    "pretty_duration",
    "pretty_func",
    "pretty_throughput",
    "rich_location",
    "save",
    "sentinel",
    "serde",
    "timer",
    "timing",
    "todo",
    "toml",
    "tqdm",
    "track",
    "try_import",
    "typing",
    "unreachable",
    "version",
    "version_tuple",
    "wrapt_getattr",
    "wrapt_setattr",
    "yaml",
]



from typing import Never

import attrs


@attrs.define
class TodoError(NotImplementedError):
    message: str = "not yet implemented"
    assignee: str | None = None

    def __str__(self) -> str:
        msg: str = "TODO"
        if self.assignee:
            msg += f"({self.assignee})"
        msg += f": {self.message}"
        return msg


@attrs.define
class UnreachableError(AssertionError):
    message: str | None = None

    def __str__(self) -> str:
        msg: str = "internal error: entered unreachable code"
        if self.message:
            msg += f": {self.message}"
        return msg


def todo(message: str = "not yet implemented", assignee: str | None = None) -> Never:
    raise TodoError(message=message, assignee=assignee)


def unreachable(message: str | None = None) -> Never:
    raise UnreachableError(message=message)



from __future__ import annotations

import loguru
from loguru import logger


def as_level_no_dict(levels: loguru.FilterDict) -> dict[str | None, int]:
    return {key: get_level_no(value) for key, value in levels.items()}


def get_level_no(level: bool | int | str) -> int:  # noqa: FBT001
    if level is True:
        return 0
    if level is False:
        return 100
    if isinstance(level, int):
        return level
    return logger.level(level).no


def get_parent(name: str) -> str:
    index: int = name.rfind(".")
    return "" if index < 0 else name[:index]



import functools
from collections.abc import Iterable
from typing import Any, overload

from ._statistics import StatisticName
from ._timings import Callback, Timings
from .defaults import (
    LOG_RECORD_DEFAULT_INDEX,
    LOG_RECORD_DEFAULT_LEVEL,
    LOG_RECORD_DEFAULT_THRESHOLD_SEC,
    LOG_SUMMARY_DEFAULT_LEVEL,
    LOG_SUMMARY_DEFAULT_STATISTICS,
)


@overload
def log_record(
    timer: Timings,
    /,
    *,
    index: int = LOG_RECORD_DEFAULT_INDEX,
    level: int | str = LOG_RECORD_DEFAULT_LEVEL,
    threshold_sec: float | None = LOG_RECORD_DEFAULT_THRESHOLD_SEC,
) -> Any: ...
@overload
def log_record(
    *,
    index: int = LOG_RECORD_DEFAULT_INDEX,
    level: int | str = LOG_RECORD_DEFAULT_LEVEL,
    threshold_sec: float | None = LOG_RECORD_DEFAULT_THRESHOLD_SEC,
) -> Callback: ...
def log_record(timer: Timings | None = None, /, **kwargs) -> Any:
    __tracebackhide__ = True
    if timer is None:
        return functools.partial(log_record, **kwargs)
    return timer.log_record(**kwargs)


@overload
def log_summary(
    timer: Timings,
    /,
    *,
    level: int | str = LOG_SUMMARY_DEFAULT_LEVEL,
    stats: Iterable[StatisticName] = LOG_SUMMARY_DEFAULT_STATISTICS,
) -> None: ...
@overload
def log_summary(
    *,
    level: int | str = LOG_SUMMARY_DEFAULT_LEVEL,
    stats: Iterable[StatisticName] = LOG_SUMMARY_DEFAULT_STATISTICS,
) -> Callback: ...
def log_summary(timer: Timings | None = None, /, **kwargs) -> Any:
    __tracebackhide__ = True
    if timer is None:
        return functools.partial(log_summary, **kwargs)
    return timer.log_summary(**kwargs)



from collections.abc import Iterable


def len_or_none(iterable: Iterable) -> int | None:
    try:
        return len(iterable)  # pyright: ignore[reportArgumentType]
    except TypeError:
        return None



import functools
import os
from collections.abc import Callable, Mapping
from typing import Any, Literal, TypedDict

import pydantic

type EncHook = Callable[[Any], Any]


type IncEx = (
    set[int] | set[str] | Mapping[int, "IncEx | bool"] | Mapping[str, "IncEx | bool"]
)


class PydanticModelDumpOptions(TypedDict, total=False):
    mode: Literal["json", "python"]
    include: IncEx | None
    exclude: IncEx | None
    context: Any | None
    by_alias: bool | None
    exclude_unset: bool
    exclude_defaults: bool
    exclude_none: bool
    round_trip: bool
    warnings: bool | Literal["none", "warn", "error"]
    fallback: Callable[[Any], Any] | None
    serialize_as_any: bool


@functools.singledispatch
def enc_hook(obj: Any, /, **_kwargs) -> Any:
    msg: str = f"Objects of type {type(obj)} are not supported"
    raise NotImplementedError(msg)


@enc_hook.register(pydantic.BaseModel)
def _(
    obj: pydantic.BaseModel,
    /,
    *,
    pydantic_options: PydanticModelDumpOptions | None = None,
    **_kwargs,
) -> Any:
    pydantic_options = pydantic_options or {}
    return obj.model_dump(**pydantic_options)


@enc_hook.register(os.PathLike)
def _(obj: os.PathLike, /, **_kwargs) -> str:
    return str(obj)



from __future__ import annotations

import contextlib
import sys
import types
from collections.abc import Hashable
from typing import Any

import attrs
import cachetools
import loguru
import packaging.version

from ._utils import get_level_no, get_parent


@attrs.define
class FilterByVersion:
    level_dev: int = attrs.field(default="TRACE", converter=get_level_no)  # pyright: ignore[reportAssignmentType]
    level_prerelease: int = attrs.field(default="DEBUG", converter=get_level_no)  # pyright: ignore[reportAssignmentType]
    _get_version_cache: cachetools.Cache[Hashable, packaging.version.Version | None] = (
        attrs.field(factory=lambda: cachetools.LRUCache(maxsize=128))
    )

    def __call__(self, record: loguru.Record) -> bool:
        level: int | None = self.get_level(record["name"])
        if level is None:
            return True
        return record["level"].no >= level

    def get_level(self, name: str | None) -> int | None:
        version: packaging.version.Version | None = self._get_version(name)
        if version is None:
            return None
        if version.is_devrelease:
            return self.level_dev
        if version.is_prerelease:
            return self.level_prerelease
        return None

    def _get_version(self, name: str | None) -> packaging.version.Version | None:
        if not name:
            return None
        module: types.ModuleType | None = sys.modules.get(name)
        if module is not None:
            version: Any = getattr(module, "__version__", None)
            if isinstance(version, tuple):
                with contextlib.suppress(TypeError):
                    version = ".".join(str(part) for part in version)
            if isinstance(version, str):
                try:
                    version = packaging.version.parse(version)
                except packaging.version.InvalidVersion:
                    pass
                else:
                    return version
        name = get_parent(name)
        return self._get_version(name)



from collections.abc import Callable, Mapping, Sequence

import attrs


@attrs.define
class Params:
    args: Sequence = ()
    kwargs: Mapping = {}


@attrs.define
class DispatchLookupError(LookupError):
    func: Callable
    params: Params

    def __init__(
        self, func: Callable, args: Sequence = (), kwargs: Mapping = {}
    ) -> None:
        params = Params(args=args, kwargs=kwargs)
        self.__attrs_init__(func=func, params=params)  # pyright: ignore[reportAttributeAccessIssue]

    def __str__(self) -> str:
        from liblaf.grapes import pretty

        pretty_call: str = pretty.pretty_call(
            self.func, self.params.args, self.params.kwargs
        )
        return f"`{pretty_call}` could not be resolved."



from collections.abc import Generator, Iterable

import wrapt

from ._base import BaseTimer
from ._utils import get_timer, set_timer


class TimedIterable[T](wrapt.ObjectProxy):
    def __init__(self, wrapped: Iterable[T], timer: BaseTimer) -> None:
        super().__init__(wrapped)
        if timer.name is None:
            timer.name = "Iterable"
        set_timer(self, timer)

    def __iter__(self) -> Generator[T]:
        __tracebackhide__ = True
        timer: BaseTimer = get_timer(self)
        timer.start()
        try:
            for item in self.__wrapped__:
                yield item
                timer.stop()
                timer.start()
        finally:
            # When the `for` loop is exhausted, it does not re-enter the loop
            # body. Therefore, the `start()` call after the *last* item is
            # redundant. However, since `timer._start_time` is not used anywhere
            # else, we can safely leave it out.
            timer.finish()



import types

import loguru
from rich.console import Console
from rich.traceback import Traceback

from liblaf.grapes import pretty

from ._traceback import rich_traceback

DEFAULT_FORMAT = (
    "<green>{elapsed}</green> "
    "<level>{level:<8}</level> "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> "
    "<level>{message}</level>"
)


def new_format(template: str = DEFAULT_FORMAT) -> "loguru.FormatFunction":
    template = template.strip() + "\n"

    def format_(record: "loguru.Record", /) -> str:
        if record["exception"] is None:
            return template
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = record["exception"]
        if exc_type is None or exc_value is None:
            return template + "{exception}\n"
        console: Console = pretty.get_console(color_system=None)
        rich_tb: Traceback = rich_traceback(exc_type, exc_value, traceback, width=128)
        with console.capture() as capture:
            console.print(rich_tb)
        record["extra"]["rich_traceback"] = capture.get()
        return template + "{extra[rich_traceback]}\n"

    return format_



from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal, overload

from liblaf.grapes.typing import PathLike

from ._encode import EncHook, PydanticModelDumpOptions
from ._serde import json, toml, yaml

writers: dict[str, Callable] = {}
writers[".json"] = json.save
writers[".toml"] = toml.save
writers[".yaml"] = yaml.save
writers[".yml"] = yaml.save


@overload
def save(  # pyright: ignore[reportInconsistentOverload]
    path: PathLike,
    obj: Any,
    /,
    *,
    enc_hook: EncHook | None = ...,
    force_ext: str | None = None,
    order: Literal["deterministic", "sorted"] | None = None,
    pydantic: PydanticModelDumpOptions | None = None,
) -> None: ...
def save(path: PathLike, obj: Any, /, force_ext: str | None = None, **kwargs) -> None:
    path = Path(path)
    ext: str = force_ext or path.suffix
    return writers[ext](path, obj, **kwargs)



from __future__ import annotations

import attrs
import loguru

from ._by_name import FilterByName
from ._by_version import FilterByVersion
from ._once import FilterOnce
from ._utils import get_level_no


@attrs.define
class CompositeFilter:
    by_name: FilterByName = attrs.field(factory=FilterByName)
    by_version: FilterByVersion = attrs.field(factory=FilterByVersion)
    level: int = attrs.field(default="INFO", converter=get_level_no)  # pyright: ignore[reportAssignmentType]
    once: FilterOnce = attrs.field(factory=FilterOnce)

    _cache_key: loguru.Record = attrs.field(default=None, init=False)
    _cache_value: bool = attrs.field(default=False, init=False)

    def __init__(self, by_name: loguru.FilterDict | None = None) -> None:
        if by_name is None:
            by_name = {"__main__": "TRACE"}
        level: bool | int | str = by_name.get("", "INFO")
        self.__attrs_init__(by_name=FilterByName(by_name), level=level)  # pyright: ignore[reportAttributeAccessIssue, reportArgumentType]

    def __call__(self, record: loguru.Record) -> bool:
        if record is self._cache_key:
            return self._cache_value
        result: bool = self._filter(record)
        self._cache_key = record
        self._cache_value = result
        return result

    def get_level(self, name: str | None) -> int | None:
        level: int | None = self.by_name.get_level(name)
        if level is not None:
            return level
        level = self.by_version.get_level(name)
        if level is not None:
            return level
        return None

    def _filter(self, record: loguru.Record) -> bool:
        if not self.once(record):
            return False
        level: int | None = self.get_level(record["name"])
        if level is None:
            level = self.level
        return record["level"].no >= level



from collections.abc import Callable, Iterable, Sequence
from typing import Any, overload

from ._clock import ClockName
from ._timer import Timer
from ._timings import Callback


@overload
def timer(
    *,
    name: str | None = ...,
    clocks: Sequence[ClockName] = ...,
    cb_finish: Callback | None = ...,
    cb_start: Callback | None = ...,
    cb_stop: Callback | None = ...,
) -> Timer: ...
@overload
def timer[C: Callable](
    callable: C,
    /,
    *,
    name: str | None = ...,
    clocks: Sequence[ClockName] = ...,
    cb_start: Callback | None = ...,
    cb_stop: Callback | None = ...,
    cb_finish: Callback | None = ...,
) -> C: ...
@overload
def timer[I: Iterable](
    iterable: I,
    /,
    *,
    name: str | None = ...,
    clocks: Sequence[ClockName] = ...,
    cb_start: Callback | None = ...,
    cb_stop: Callback | None = ...,
    cb_finish: Callback | None = ...,
) -> I: ...
def timer(func_or_iterable: Callable | Iterable | None = None, /, **kwargs) -> Any:
    timer = Timer(**kwargs)
    if func_or_iterable is None:
        return timer
    return timer(func_or_iterable)



import contextlib
import contextvars
import types
from collections.abc import Callable
from typing import Any, Self, overload, override

import attrs
import wrapt
from typing_extensions import deprecated

import liblaf.grapes.itertools as it

_depth: contextvars.ContextVar[int] = contextvars.ContextVar("depth", default=0)


@attrs.define
class DepthTrackerDecorator(contextlib.AbstractContextManager):
    _depth_inc: int | None = attrs.field(default=None, alias="depth_inc")
    _token: contextvars.Token[int] | None = attrs.field(default=None, init=False)

    @override  # impl contextlib.AbstractContextManager
    def __enter__(self) -> Self:
        depth_inc: int = it.first_not_none(self._depth_inc, 1)
        self._token = _depth.set(_depth.get() + depth_inc)
        return self

    @override  # impl contextlib.AbstractContextManager
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        assert self._token is not None
        _depth.reset(self._token)
        del self._token

    def __call__[**P, T](self, func: Callable[P, T], /) -> Callable[P, T]:
        @wrapt.decorator
        def wrapper(
            wrapped: Callable[P, T],
            _instance: Any,
            args: tuple[Any, ...],
            kwargs: dict[str, Any],
        ) -> T:
            __tracebackhide__ = True
            depth_inc: int = it.first_not_none(self._depth_inc, 2)
            token: contextvars.Token[int] = _depth.set(_depth.get() + depth_inc)
            try:
                return wrapped(*args, **kwargs)
            finally:
                _depth.reset(token)

        return wrapper(func)


@attrs.define
class DepthTracker:
    @overload
    def __call__(self, /, *, depth: int | None = None) -> DepthTrackerDecorator: ...
    @overload
    def __call__[C: Callable](self, func: C, /, *, depth: int | None = None) -> C: ...
    @deprecated("Use `__tracebackhide__ = True` instead.")
    def __call__(
        self, func: Callable | None = None, /, *, depth: int | None = None
    ) -> Callable:
        decorator = DepthTrackerDecorator(depth_inc=depth)
        if func is None:
            return decorator
        return decorator(func)

    @property
    @deprecated("Use `__tracebackhide__ = True` instead.")
    def depth(self) -> int:
        return _depth.get()


helper: DepthTracker = DepthTracker()
